<?php defined('SYSPATH') or die('No direct script access.');

class Model_Rating extends Model
{
    public static function GetStudentsForRating($discipline) {
        $sql = 'CALL `GetStudentsForRating`(:discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->execute();
    }

    /**
     * @param $discipline int discipline id
     * @param $student    int student id
     * @return Database_Result
     */
    public static function getRates($discipline, $student) {
        $sql = 'CALL `GetRates`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }

    /**
     * @param $discipline int discipline id
     * @param $student    int student id
     * @return Database_Result
     */
    public static function getRatesAll($discipline, $student) {
        $sql = 'CALL `GetRatesAll`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }


    public static function getMapForStudent($student, $discipline) {
        $sql = 'CALL `GetMapForStudent`(:student, :discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':student', $student)
            ->execute();
    }

    public static function getAttestationData($discipline, $group) {
        $sql = 'CALL `GetAttestationData`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->param(':group', $group)
            ->execute();
    }

    private static function checkRateChanges($studentId, $disciplineId, $submoduleId, $rate) {
        $studentsRates = Model_Rating::getRatesAll($disciplineId, $studentId);
//        echo Debug::vars($studentsRates);

        $change = ['type' => '', 'rate' => 0];
        $ratesByType = [];
        function getKey($part1, $part2) {
            return $part1 == 'exam' ? $part1.$part2 : $part1;
        }

        foreach ($studentsRates as $rateInfo) {
            $key = getKey($rateInfo['ModuleType'], $rateInfo['SubmoduleOrderNum']);

            // запоминаем оценки по типам
            $iRate = (int)$rateInfo['Rate'];
            if (is_null($ratesByType[$rateInfo['ModuleType']])) {
                if ($rateInfo['ExamPeriodOption'] == 'absence') {
                    $ratesByType[$key] = 0;
                }
                elseif ($rateInfo['ExamPeriodOption'] == 'pass') {
                    $ratesByType[$key] = 40;
                }
                else
                    $ratesByType[$key] = $rateInfo['Rate'] == null ? null : $iRate;
            }
            else {
                $ratesByType[$key] += $iRate;
            }

            // запомним, где пользователь внес изменения
            if ((int)$submoduleId == (int)$rateInfo['SubmoduleID']) {
                $change['type'] = $key;
                $change['rate'] = (int)$rate - $iRate;
            }
        }

        return RateStatesModel::checkTransitionForExam($ratesByType, $change);
    }

    public static function SetStudentRate($teacherId, $studentId, $disciplineId, $submoduleId, $rate) {

        $errorMsg = Model_Rating::checkRateChanges($studentId, $disciplineId, $submoduleId, $rate);
        if ($errorMsg !== '') {
            $result['success'] = false;
            $result['type'] = 'check';
            $result['errorMsg'] = $errorMsg;
            return $result;
        }

        $sql = 'SELECT `SetStudentRate`(:teacher, :student, :submodule, :rate) AS `Num`';
        $dbAnswer = DB::query(Database::SELECT, $sql)
            ->parameters([
                ':teacher'   => $teacherId,
                ':student'   => $studentId,
                ':submodule' => $submoduleId,
                ':rate'      => (int) $rate
            ])->execute();

        $result['success'] = $dbAnswer->get('Num') == 0;
        if (!$result['success']) {
            $result['type'] = 'db';
            $result['errorMsg'] = 'Ошибка приложения!';
        }
        return $result;
    }

    public static function getRatesForDiscipline(Model_Discipline $discipline) {
        $sql = 'CALL `GetRatesForDiscipline`(:id)';
        $result = DB::query(Database::SELECT, $sql)
            ->param(':id', $discipline->ID)->execute();

        $table = [];
        foreach ($result as $row) {
            $student = $row['StudentID'];
            $submodule = $row['SubmoduleID'];
            $table[$student][$submodule] = $row['Rate'];
        }

        return $table;
    }

    public static function getRatesForStudentsGroup($discipline, $group) {
        $sql = 'CALL `GetRatesForGroup`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
            ])->execute();
    }

    public static function getRatesForStudentsGroupByStage($discipline, $group, $stage) {
        $sql = 'CALL `GetRatesForGroupByStage`(:discipline, :group, :stage)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
                ':stage'      => $stage
            ])->execute();
    }

    /** @return  Database_RESULT   Database_Result for SELECT queries */
    public static function getRatesForGroupAll($group, $semester) {
        $sql = 'CALL `GetRatesForGroupAll`(:group, :semester)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':group'    => $group,
                ':semester' => $semester
            ])->execute();
    }

    public static function getFinalFormInfo($discipline, $group) {
        $sql = 'CALL `getFinalFormInfo`(:discipline, :group)';
        return DB::query(Database::SELECT, $sql)
            ->parameters([
                ':discipline' => $discipline,
                ':group'      => $group,
            ])->execute()[0];
    }

    /**
     * Get count of related with discipline records in `rating_table`.
     * @param Model_Discipline $discipline
     * @return int
     */
    public static function count(Model_Discipline $discipline) {
        $sql = 'SELECT `Discipline_CountRatings`(:discipline) AS `res`';
        $res = DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline->ID)
            ->execute();

        return (int) $res->get('res');
    }

    public static function getHistory($discipline) {
        $sql = 'CALL `GetRatesHistory`(:discipline)';
        return DB::query(Database::SELECT, $sql)
            ->param(':discipline', $discipline)
            ->execute();
    }

    public static function setExamPeriodOption($studentId, $disciplineId, $submoduleId, $option) {
        $rateForCheckOnly = '';
        if ($option == 'null' or $option == null) {
            $rateForCheckOnly = '';
        }
        elseif ($option == 'absence') {
            $rateForCheckOnly = '0'; // не сдал, а баллы только для проверки
        }
        elseif ($option == 'pass') {
            $rateForCheckOnly = '40'; // сдал, а баллы только для проверки
        }
        else {
            return ['success' => false, 'type' => 'check', 'errorMsg' => 'Unexpected option'];
        }

        $errorMsg = Model_Rating::checkRateChanges($studentId, $disciplineId, $submoduleId, $rateForCheckOnly);
        if ($errorMsg !== '') {
            return ['success' => false, 'type' => 'check', 'errorMsg' => $errorMsg];
        }

        // option is absence or pass or null
        $sql = 'SELECT `Discipline_SetExamPeriodOption`(:student, :submodule, :option) as `ErrorCode`';
        $result = DB::query(Database::SELECT, $sql)
            ->parameters([
                ':student'   => $studentId,
                ':submodule' => $submoduleId,
                ':option'    => $option
            ])->execute()->get('ErrorCode') == 0;
        return ['success' => $result, 'type' => 'db request',
            'errorMsg' => 'Ошибка БД! Выставления опции экзамена не удалось!'];
    }
};

class States
{
    const __default = self::beforeLimit;

    const beforeLimit = 0; // не достигнут порог 38
    const semesterAccepted = 1;     // допущен к сдаче
    const notPassed1 = 2;  // не сдал один раз (завалил основную сдачу)
    const notPassed2 = 3;  // не сдал дважды (завалил первую пересдачу)
    const notPassed3 = 4;  // не сдал трижды (завалил вторую пересдачу)
    const passed1 = 5;      // сдал экзамен в основную сдачу
    const passed2 = 6;      // сдал экзамен со 2-го раза
    const passed3 = 7;      // сдал экзамен с 3-го раза
};

class Transition
{
    public $state = null;
    public $newState = null;
    public $description = null;

    function __construct($state, $newState, $description) {
        $this->state = $state;
        $this->newState = $newState;
        $this->description = $description;
    }
};

class RateStatesModel
{
    protected $forward = null;
    protected $backward = null;

    // Список некорректных переходов содержит тройки вида: Текущее Состояние, Следующее Состояние, Сообщение
    // об Ошибке.
    // Текущее Состояние - состояние, в котором находятся баллы студента к моменту внесения изменений.
    // Следующее Состояние - состояние, в котором будут находится баллы студента после принятия изменений.
    // Сообщение об Ошибке - описание ошибки которое будет продемонстировано преподавателю.
    protected $incorrectTransitions = [ 'inFuture' => [], 'inPast' => []];

    function __construct()
    {
        $this->incorrectTransitions["inFuture"] = [
            new Transition(States::beforeLimit, States::notPassed1, "Недостаточно баллов для допуска к сдаче экзамена!"),
            new Transition(States::beforeLimit, States::notPassed1, 'Недостаточно баллов для допуска к сдаче экзамена!'),
            new Transition(States::beforeLimit, States::notPassed2, 'Недостаточно баллов для допуска к пересдаче '.
                'экзамена!'),
            new Transition(States::beforeLimit, States::notPassed3, 'Недостаточно баллов для допуска к пересдаче '.
                'экзамена!'),
            new Transition(States::beforeLimit, States::passed1, 'Недостаточно баллов для допуска к сдаче экзамена!'),
            new Transition(States::beforeLimit, States::passed2, 'Недостаточно баллов для допуска к сдаче экзамена!'),
            new Transition(States::beforeLimit, States::passed3, 'Недостаточно баллов для допуска к сдаче экзамена!'),
            new Transition(States::semesterAccepted, States::notPassed2, 'Не выставлены баллы за основную сдачу!'),
            new Transition(States::semesterAccepted, States::notPassed3, 'Не выставлены баллы за основную сдачу!'),
            new Transition(States::semesterAccepted, States::passed2, 'Не выставлены баллы за основную сдачу!'),
            new Transition(States::semesterAccepted, States::passed3, 'Не выставлены баллы за основную сдачу!'),
            new Transition(States::notPassed1, States::notPassed3, 'Не выставлены баллы за первую пересдачу!'),
            new Transition(States::notPassed1, States::passed3, 'Не выставлены баллы за первую пересдачу!'),
        ];

        $this->incorrectTransitions["inPast"] = [
            new Transition(States::notPassed1, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за экзамен!'),
            new Transition(States::notPassed2, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за первую пересдачу!'),
            new Transition(States::notPassed2, States::semesterAccepted, 'Нельзя удалять баллы за основную сдачу, при '.
                'наличии баллов за пересдачу!'),
            new Transition(States::notPassed2, States::passed1, 'Нельзя увеличивать баллы выше порога за основную ' .
                'сдачу, при наличии баллов за первую пересдачу!'),
            new Transition(States::notPassed3, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за экзамен!'),
            new Transition(States::notPassed3, States::semesterAccepted, 'Недопустимое изменение баллов!'),
            new Transition(States::notPassed3, States::notPassed1, 'Недопустимое изменение баллов!'),
            new Transition(States::notPassed3, States::passed1, 'Недопустимое изменение баллов!'),
            new Transition(States::notPassed3, States::passed2, 'Нельзя увеличивать баллы выше порога за первую ' .
                'пересдачу, при наличии баллов за вторую пересдачу!'),
            new Transition(States::passed1, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за экзамен!'),
            new Transition(States::passed2, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за первую пересдачу!'),
            new Transition(States::passed2, States::semesterAccepted, 'Нельзя удалять баллы за основную сдачу, при '.
                'наличии баллов за пересдачу!'),
            new Transition(States::passed2, States::passed1, 'Нельзя увеличивать баллы выше порога за основную ' .
                'сдачу, при наличии баллов за первую пересдачу!'),
            new Transition(States::passed3, States::beforeLimit, 'Нельзя уменьшать баллы в семестре ниже порога, '.
                'при наличии неявки или баллов за экзамен!'),
            new Transition(States::passed3, States::semesterAccepted, 'Нельзя удалять баллы за основную сдачу, при '.
                'наличии баллов за пересдачу!'),
            new Transition(States::passed3, States::passed1, 'Нельзя увеличивать баллы выше порога за основную ' .
                'сдачу, при наличии баллов за вторую пересдачу!'),
            new Transition(States::passed3, States::passed2, 'Нельзя увеличивать баллы выше порога за первую ' .
                'пересдачу, при наличии баллов за вторую пересдачу!'),
        ];
    }

    private function fillForExam($semester, $gain, $examAttempts) {
//        echo Debug::vars($examAttempts);

        $iSemester = (int)$semester;
        $iGain = (int)$gain; // преобразует в int и null преобразуется в 0.
        $iExamAttempts = array_map('intval', $examAttempts);

        if ($iSemester + $iGain < 38) {
            $this->forward = States::beforeLimit;
        }
        else if ($examAttempts[0] === null) {
            $this->forward = States::semesterAccepted;
        }
        else if ($iExamAttempts[0] >= 22) {
            $this->forward = States::passed1;
        }
        else if ($examAttempts[1] === null) {
            $this->forward = States::notPassed1;
        }
        else if ($iExamAttempts[1] >= 22) {
            $this->forward = States::passed2;
        }
        else if ($examAttempts[2] === null) {
            $this->forward = States::notPassed2;
        }
        else if ($iExamAttempts[2] >= 22) {
            $this->forward = States::passed3;
        }
        else {
            $this->forward = States::notPassed3;
        }

        if ($examAttempts[2] !== null) {
            if ($iExamAttempts[2] < 22) {
                $this->backward = States::notPassed3;
            }
            else {
                $this->backward = States::passed3;
            }
        }
        else if ($examAttempts[1] !== null) {
            if ($iExamAttempts[1] < 22) {
                $this->backward = States::notPassed2;
            }
            else {
                $this->backward = States::passed2;
            }
        }
        else if ($examAttempts[0] !== null) {
            if ($iExamAttempts[0] < 22) {
                $this->backward = States::notPassed1;
            }
            else {
                $this->backward = States::passed1;
            }
        }
        else if ($iSemester + $iGain < 38) {
            $this->backward = States::beforeLimit;
        }
        else {
            $this->backward = States::semesterAccepted;
        }
    }

    private function checkTransitionPermission($previousState) {

        function search($where, $what, $previousState) {
            $res = '';
            foreach ($where as $transition) {
                if (($previousState === $transition->state) && ($what === $transition->newState)) {
                    $res = $transition->description;
                    break;
                }
            }
            return $res;
        }

        $res1 = search($this->incorrectTransitions["inFuture"], $this->backward, $previousState);
        $res2 = search($this->incorrectTransitions["inPast"], $this->forward, $previousState);
        return $res1.$res2;
    }

    public static function checkTransitionForExam($ratesByType, $change) { // ($previousState, $semester, $gain='', $examAttempts) {
        // определить предыдущее состояние
        $model = new RateStatesModel();
        $model->fillForExam($ratesByType['regular'], $ratesByType['extra'],
            [$ratesByType['exam1'], $ratesByType['exam2'], $ratesByType['exam3']]);
        $previousState = $model->forward;
//        echo Debug::vars($ratesByType);
//        echo Debug::vars('Forward: ', $model);

        // определить текущее состояние
        $ratesByType[$change['type']] += $change['rate'];
        if ($ratesByType[$change['type']] == -1) {
            $ratesByType[$change['type']] = null;
        }
        $model->fillForExam($ratesByType['regular'], $ratesByType['extra'],
            [$ratesByType['exam1'], $ratesByType['exam2'], $ratesByType['exam3']]);
//        echo Debug::vars('Backward: ', $model);
        // анализ ошибок
        return $model->checkTransitionPermission($previousState);
    }
};
